# -*- coding: utf-8 -*-
# contacts.models
import datetime

from django.db import models
from django.core import exceptions
from django.core.urlresolvers import reverse
import django.contrib.auth.models as auth_models
from django.core.exceptions import ValidationError

import references.models as ref
import contacts.forms as contacts_forms
import contacts.fields_choices as fch
from contacts.logger import logger
from contacts import settings_messages as msg

LOCAL_DATE_FORMAT = '%d/%m/%Y'

### Utilitaires
def get_rem_user():
    return auth_models.User.objects.get_or_create(username='deleted')[0]


def no_attrib_error(orig_func):
    "return empty unicode if attrib does not exist"
    def wrapper(*args, **kwargs):
        try:
            return orig_func(*args, **kwargs)
        except AttributeError:
            return u""
    return wrapper


### Classes principales
class AdhManager(auth_models.UserManager):
    def get_query_set(self):
        return super(AdhManager, self).get_query_set().filter(
            groups__isnull=True, is_superuser=False,)


class AdhLogin(auth_models.User):
    class Meta:
        proxy = True
        verbose_name = u"Compte adhérent"
        verbose_name_plural = u"      Comptes adhérents"
    objects_adh = AdhManager()
    objects = auth_models.UserManager()

    def is_adherent(self):
        "poor man test: si ne fait partie de aucun groupe, c'est un adherent"
        return bool(self.groups)


class Contact(object):
    """
    Classe d'interface définissant l'API commune entre les modèles Structure et
    Personne qui sont parfois traités de façon identique puisqu'ils partagent
    le même panier.
    """
    # Champs à inclure dans un export CSV pour publipostage
    attrs_publipostage = [
            'get_designation', 'adresse', 'code_postal', 'ville',
            'tel_fixe', 'tel_mobile', 'fax', 'courriel', ]

    def get_designation(self):
        """
        Retourne le nom de la structure ou de la première personne associée.
        (utilisée lors de l'exportation CSV pour publipostage)
        """
        return unicode(self)
    get_designation.short_description = u"Désignation ou personne"

    def adresse_electronique(self):
        """
        Retourne une liste de 2 éléments:
        - le nom de la structure
        - l'adresses email associée
        Retourne une liste vide si aucune adresse courriel n'est renseignée.
        """

    def get_courriels(self):
        """
        Retourne le courriel de la structure et des personnes associées
        (sauf s'ils ne souhaitent pas recevoir d'email automatique)
        """

    def liste_contacts_associes(self, courrier=False):
        """
        Méthode commune aux modèles Structure et Personne pour un traitement
        indifférencié de ces objets. Retourne la liste des contacts associés à
        cet objet. Pour un objet Personne, il n'y en a pas.
        Si courrier vaut True, retourne uniquement ceux devant recevoir un
        courrier destiné à la structure (utilisé uniquement dans Structure).
        """
        return []


class Structure(Contact, models.Model):
    u"""
    Personnes Morales

    Attributs et méthodes hérités de la classe abstraite Contact (hérités aussi
    par Personne) pour permettre le partage du même panier avec le modèle
    Personne.
     - attrs_publipostage
     - adresse_electronique()
     - get_courriels()
    """
    class Meta:
        ordering = ['designation',]
        verbose_name = "Personne Morale"
        verbose_name_plural = "         Personnes Morales"

    # Champs à inclure dans un export CSV complet
    attrs_export = (
        'id',
        'type',
        'f_juridique',
        'get_designation',
        'personne_set',
        'get_youngest_person__date_naissance',
        'siret',
        'statut',
        'adresse',
        'ville',
        'code_postal',
        'commune__code_commune',
        'commune',
        'commune__arrondissement',
        'commune__zone',
        'commune__enjeu_eau',
        'tel_fixe',
        'tel_mobile',
        'fax',
        'courriel',
        'site_web',
        'date_creation',
        'date_derniere_modification',
        'utilisateur_derniere_modification',
        'origine_contact',
        'annee_contact',
        'mois_contact',
        'commentaire',
        'etiquettes',
        'visibilite',
        'type_activite_pri',
        'qual_compl_conv',
        'type_activite',
        'qual_compl',
        'last_donneesbio__ass_total',
        'last_donneesbio__ass_total_bio',
        'sau_eau',
        'sau_eau_bio',
        'premiere_adhesion',
        'get_date_derniere_adhesion',
        'get_date_premiere_conversion',
        )
    adherent = models.ForeignKey(
        AdhLogin, null=True, blank=True, on_delete=models.SET_NULL,
        verbose_name=u"compte adhérent", help_text=
        u"compte permettant la connexion à l'application")
    designation = models.CharField(u"désignation", max_length=250)
    f_juridique = models.ForeignKey(
        ref.FormeJuridique, null=True, blank=True, on_delete=models.PROTECT,
        verbose_name=u"Forme juridique")
    statut = models.ForeignKey(
        ref.StatutStructure, null=True, blank=True, on_delete=models.SET_NULL)
    adresse = models.TextField(null=True, blank=True)
    commune = models.ForeignKey(
        ref.Commune, null=True, blank=True, on_delete=models.PROTECT)
    # Certaines communes peuvent avoir plusieurs codes postaux
    code_postal = models.CharField(
        max_length=10, null=True, blank=True,
        help_text=u"Laisser vide pour utiliser celui associé à la commune")
    # la ville contient également le code cedex (à utilser pour l'adresse)
    ville = models.CharField(
        max_length="255", null=True, blank=True,
        help_text=u"Nom de ville avec code cédex si nécessaire, "
        u" si différent du nom de la commune.")
    tel_fixe = models.CharField(
        u"téléphone fixe", max_length=100, null=True, blank=True)
    tel_mobile = models.CharField(
        u"téléphone mobile", max_length=100, null=True, blank=True)
    fax = models.CharField(max_length=100, null=True, blank=True)
    courriel = models.CharField(max_length=250, null=True, blank=True)
    siret = models.CharField(
        u"N. Siret", max_length=14, null=True, blank=True,
        help_text="14 chiffres sans espaces")
    pas_de_courriel = models.BooleanField(
        u"pas d'envoi automatique de courriel", default=False,
        help_text=u"Cochez cette case si cette structure ne souhaite"
        u" pas recevoir de courriel envoyé automatiquement.")
    site_web = models.URLField(
        u"URL du site web", max_length=250, null=True, blank=True)
    date_creation = models.DateField(
        u"date de saisie de l'enregistrement", editable=False)
    date_derniere_modification = models.DateField(
        u"date de dernière modification des informations", editable=False)
    utilisateur_derniere_modification = models.ForeignKey(
        auth_models.User, editable=False, on_delete=models.PROTECT,
        verbose_name=u"dernière modification par")
    origine_contact = models.CharField(
        u"origine du contact", max_length=255, null=True, blank=True)
    annee_contact = models.IntegerField(
        u"Année du premier contact", null=True, blank=True,
        choices=fch.CHOIX_ANNEE)
    mois_contact = models.IntegerField(
        u"Mois du premier contact", null=True, blank=True,
        choices=fch.CHOIX_MOIS)
    sage = models.CharField(u"SAGE", max_length=255, null=True, blank=True)
    commentaire = models.TextField(null=True, blank=True)
    etiquettes = models.ManyToManyField(
        ref.Etiquette, null=True, blank=True, verbose_name=u'étiquettes')
    type_pm = models.IntegerField(
        u"Type personne morale", default=0, choices=fch.TYPE_PM)
    # from agriculteur
    visibilite = models.CharField(
        max_length=255, choices=fch.CHOIX_VISIBILITE, default="personne",
        verbose_name=u"publication sur le site gabnor.org",)
    type_activite = models.ForeignKey(
        ref.TypeActivite, null=True, blank=True, on_delete=models.PROTECT,
        verbose_name=u"Activité principale bio")
    type_activite_pri = models.ForeignKey(
        ref.TypeActivite, null=True, blank=True, related_name="structure_pri",
        on_delete=models.PROTECT, verbose_name=u"Activité principale")
    qual_compl = models.ManyToManyField(
        ref.AutreQualificatifExploitation, null=True, blank=True,
        verbose_name=u"autres activités bio")
    qual_compl_conv = models.ManyToManyField(
        ref.AutreQualificatifExploitation, null=True, blank=True,
        verbose_name=u"autres activités conventionnelles",
        related_name="structure_conv")
    # si commune.enjeu_eau
    sau_eau = models.DecimalField(
        max_digits=10, decimal_places=2, null=True, blank=True,
        verbose_name=u"SAU à enjeu eau",
        help_text='Surface agricole utile sur commune à enjeu eau (en ha)')
    sau_eau_bio = models.DecimalField(
        max_digits=10, decimal_places=2, null=True, blank=True,
        verbose_name=u"SAU en biologique à enjeu eau",
        help_text=u'Surface agricole utile en biologique'
        u' sur commune à enjeu eau (en ha)')
    # fin si commune.enjeu_eau
    parc_naturel_regional = models.CharField(
        max_length=200, verbose_name=u"parc naturel régional",
        choices=fch.CHOIX_PARCS_NATURELS_REGIONAUX, null=True, blank=True)
    # from agriculteurbio
    premiere_adhesion = models.ForeignKey(
        'AdhesionGabnor', null=True, blank=True, on_delete=models.SET_NULL,
        related_name="premieres_adhesions", verbose_name=u"Première adhésion",)
    renouvellement_adhesion = models.ForeignKey(
        'AdhesionGabnor', null=True, blank=True,
        related_name="renouvellements", on_delete=models.SET_NULL,
        verbose_name=u"dernier renouvellement d'adhésion")

    def __unicode__(self):
        "Retourne une représentation unicode de la structure."
        if self.designation:
            return self.designation
        if self.personne_set.count() > 0:
            return unicode(self.personne_set.all()[0])
        return u"Structure (#%d)" % self.id

    def get_absolute_url(self):
        "Retourne l'URL associée à l'objet."
        return reverse(
            'admin:{0}_{1}_change'.format(
                self._meta.app_label, self.__class__.__name__.lower()),
            args=[self.pk])

    def contenu_etiquette(self):
        """
        Retourne le contenu à afficher dans l'étiquette sous forme de liste
        dont chaque élément représente une ligne.
        """
        contenu = []
        # On doit remplacer & par &amp; à cause d'un problème dans reportlab
        # qui génère une erreur XML lorsque le nom contient un "&"
        contenu.append(unicode(self).replace("&", "&amp;"))
        contenu.extend(self.adresse.split('\n'))
        contenu.append(u"%s %s" % (self.code_postal or '', self.ville or ''))
        return contenu

    def liste_contacts_associes(self):
        """
        Surcharge de la méthode de Contact.
        Retourne la liste des contacts associés à cet objet. Il
        s'agit ici des personnes associées à la structure.
        """
        return self.personne_set.all()

    def adresse_electronique(self):
        """
        Retourne une liste de 2 éléments:
        - le nom de la structure
        - l'adresse email associée
        Retourne une liste vide si aucune adresse courriel n'est renseignée.
        """
        if not self.courriel:
            return ()
        return (unicode(self), self.courriel)

    def get_courriels(self):
        """
        Retourne le courriel de la structure et des personnes associées
        (sauf s'ils ne souhaitent pas recevoir d'email automatique)
        """
        courriels = []
        if self.courriel:
            courriels.append(self.adresse_electronique())
        for p in self.liste_contacts_associes():
            courriels.extend(p.get_courriels())
        return courriels

    def type(self):
        "retourne le type parmi structure, agri, agribio"
        return self.get_type_pm_display()

    def clean(self):
        if self.type_activite and self.type_pm != 2:
            raise ValidationError(msg.ERROR_MESSAGE["acti_bio_sans_agri_bio"])
        if self.type_pm == 2 and not self.type_activite:
            raise ValidationError(msg.ERROR_MESSAGE["agri_bio_sans_acti_bio"])
        if self.type_activite_pri and self.type_pm == 0:
            raise ValidationError(msg.ERROR_MESSAGE["acti_pri_sans_agri"])
        if (self.type_pm >= 1 and not
            (self.type_activite or self.type_activite_pri)):
            raise ValidationError(msg.ERROR_MESSAGE["agri_sans_acti"])
        if self.f_juridique == None:
            raise ValidationError(msg.ERROR_MESSAGE["pas_f_juridique"])

    def save(self, *args, **kwargs):
        """
        Remplit les champs date_creation et date_derniere_modification
        manuellement (la fonctionnalité auto_now de django est douteuse,
        mieux vaut le gérer manuellement).
        """
        rawsave = kwargs.pop('rawsave', False)
        if rawsave:
            models.Model.save(self, *args, **kwargs)
        if self.id is None and self.date_creation is None:
            self.date_creation = datetime.date.today()
        self.date_derniere_modification = datetime.date.today()
        models.Model.save(self, *args, **kwargs)

    def get_date_premiere_adhesion(self):
        """
        Retourne la date de première adhésion, ou None s'il n'est pas abonné.
        """
        try:
            return self.premiere_adhesion.date
        except AdhesionGabnor.DoesNotExist:
            return None
    get_date_premiere_adhesion.short_description = u"date de première adhésion"

    def get_date_derniere_adhesion(self):
        """
        Retourne la date de dernière adhésion, ou None s'il n'est pas abonné.
        """
        try:
            return self.adhesiongabnor_set.latest().date
        except AdhesionGabnor.DoesNotExist:
            return None
    get_date_derniere_adhesion.short_description = u"date de dernière adhésion"

    def get_date_premiere_conversion(self):
        """
        Retourne la date de première conversion bio, ou None s'il n'y en a pas
        (ce qui ne devrait pas être le cas pour un agriculteur bio, mais
        restons parano^Wprudent).
        """
        conv = (self.projetbio_set.filter(date_engagement_oc__isnull=False)
                                 .order_by('date_engagement_oc'))
        if len(conv) == 0:
            return None
        return conv[0].date_engagement_oc
    get_date_premiere_conversion.verbose_name = u"date de première conversion"
    get_date_premiere_conversion.short_description = (
        u"date de première conversion")

    def premiere_conversion(self):
        return self.get_date_premiere_conversion()

    def last_donneesbio(self):
        if not hasattr(self, '_last_dbio'):
            try:
                self._last_dbio = self.donneesbio_set.order_by('-annee')[:1][0]
            except IndexError:
                self._last_dbio = None
        return self._last_dbio
    last_donneesbio.target_model = 'DonneesBio'

    def get_liste_donneesbio(self):
        """
        Retourne la listes des données de l'observatoire de l'agriculteur
        (utilisé pour afficher les liens vers ceux-ci dans la vue de
        modification de l'agriculteur).
        """
        return [
            {
                'url': dbio.get_absolute_url(),
                'annee': dbio.annee,
            }
            for dbio in self.donneesbio_set.order_by('-annee')
            ]

    def get_url_cree_donneesbio(self):
        """
        Retourne l'URL de la vue permettant de créer un nouvel objet DonneesBio
        pour cet agriculteur et d'afficher sa page de modification.
        """
        return reverse('admin:cree_donneesbio', args=[self.id])

    def get_youngest_person(self):
        "return youngest person attached to this structure if any"
        try:
            return self.personne_set.latest('date_naissance')
        except models.ObjectDoesNotExist:
            return None
    get_youngest_person.target_model = 'Personne'
    
    def get_rdv_coll(self):
        rdv_liste = []
        rdvP_liste = (
            RdvPersonne.objects.filter(personne__structure_id=self.id)
            .order_by('-rdv__date')[:10])
        return rdvP_liste
        for rdvP in rdvP_liste:
            if rdvP.rdv not in rdv_liste:
                rdv_liste.append(rdvP.rdv)
        return rdv_liste

    def get_rdv_ind(self):
        return self.rdv_ind_set.all()


class StructureRo(Structure):
    "La version read-only de structure"
    class Meta:
        proxy = True
        verbose_name = "Personne Morale (lecture seule)"
        verbose_name_plural = "       Personnes Morales (lecture seule)"


class ProjetBio(models.Model):
    """
    Projet de passage en bio d'un agriculteur.
    Chaque agriculteur peut en avoir plusieurs, à conserver même après le
    passage en Bio.
    """
    class Meta:
        verbose_name = "projet de passage en bio"
        verbose_name_plural = "      Projets de passage en bio"

    # Champs à inclure dans un export CSV complet
    attrs_export = [
        'id', 'agriculteur', 'agri_id', 'technicien', 'description',
        'freins_identifies', 'suites_a_donner', 'remarques',
        'etude_demandee', 'date_etude', 'date_probable_certification',
        'date_certification', 'engagement_surf', 'engagement_oc_enjeu_eau',
        'installation_bio', 'date_engagement_oc', 'commentaire', ]

    agriculteur = models.ForeignKey('Structure', on_delete=models.PROTECT)
    technicien = models.ForeignKey(
        auth_models.User, null=True, blank=True, verbose_name=u"Suivi par",
        on_delete=models.SET_NULL)
    description = models.TextField(
        null=True, blank=True, verbose_name=u"description")
    freins_identifies = models.TextField(
        u"Points favorables et freins à la conversion",
        null=True, blank=True)

    pond_savoirfaire = models.IntegerField(
        u"Savoir faire", default=0, choices=fch.POND_CONV, help_text=
        u"Manque de savoir-faire ou de connaissances (projet mal défini...),"
        u" peur de ne pas être à la hauteur, manque de rigueur, manque de "
        u"qualification de l'exploitant, formation initiale non adaptée")
    pond_motivation = models.IntegerField(
        u"Motivation", default=0, choices=fch.POND_CONV, help_text=
        u"Manque de motivation, faible sensibilité à la bio, peu enclin au "
        u"changement, proche de la retraite, pas envie de réorienter un "
        u"système intensif, pas de frein identifié, mauvaise image de la bio,"
        u" peur de perte de revenu non-avérée.")
    pond_entourage = models.IntegerField(
        u"Entourage", default=0, choices=fch.POND_CONV, help_text=
        u"Pression de l'entourage (familles, conseillers techniques ou "
        u"financier, système intégré...)")
    pond_foncier = models.IntegerField(
        u"Foncier", default=0, choices=fch.POND_CONV, help_text=
        u"Problème de foncier (manque de surface, menace de perdre des "
        u"surfaces, manque de surface accessible aux vaches, qualité des "
        u"terres...), pas de foncier dans le cas d'une installation")
    pond_materiel = models.IntegerField(
        u"Matériel", default=0, choices=fch.POND_CONV, help_text=
        u"Moyens de productions faibles ou inadaptés et difficilement "
        u"adaptables, génétique animale ou végétale peu adaptée, race de "
        u"vache, variété de pomme, manque de main d'œuvre, bâtiment inadapté"
        u" trop petit ou pas aux normes, bâtiment de stockage inadapté")
    pond_reglementation = models.IntegerField(
        u"Réglementation", default=0, choices=fch.POND_CONV, help_text=
        u"Problème réglementaires, mixité, engagé dans un cahier des charges,"
        u" forfait fiscal, règle ICPE, MAEt ou PAE.")
    pond_commercialisation = models.IntegerField(
        u"Commercialisation", default=0, choices=fch.POND_CONV, help_text=
        u"Commercialisation, difficulté à passer des hausses de prix en vente"
        u" directe, produit sans valorisation bio ou sans filière longue "
        u"(betterave sucrières, moutons, volaille de chair), plus value jugée"
        u" faible en bio, perte de revenu, cahier des charges privé")
    pond_financement = models.IntegerField(
        u"Financement", default=0, choices=fch.POND_CONV, help_text=
        u"Manque de financement du projet, manque d'aides financières (Peu ou"
        u"pas d'aide à la conversion), dans le cas des installations : peu ou"
        u" pas d'apport de capital/incapacité à investir/pas d'accès aux aides"
        u" installation")
    pond_maindoeuvre = models.IntegerField(
        u"Main d'Oeuvre ", default=0, choices=fch.POND_CONV, help_text=
        u"Manque de qualification de la main d'œuvre à embaucher, Manque de "
        u"main d'œuvre")
    suites_a_donner = models.TextField(
        null=True, blank=True, verbose_name=u"suites à donner")
    remarques = models.TextField(
        null=True, blank=True, verbose_name=u"remarques")
    etude_demandee = models.NullBooleanField(u"Étude demandée ?")
    date_etude = models.DateField(u"Date de l'étude", null=True, blank=True)
    date_probable_certification = models.CharField(
        u"date probable de début de conversion", null=True, blank=True,
        max_length=20, choices=fch.CHOIX_TERME, help_text=u"Si vous "
        u"choisissez l'option « Date précisée ci-après », merci de préciser"
        u" la date.")
    date_certification = models.DateField(
        u"Date de début de conversion prévue", null=True, blank=True,
        help_text=u"à remplir si vous avez choisi l'option \"Date précisée\"")
    engagement_surf = models.DecimalField(
        u"Surface engagée (en ha)",
        null=True, blank=True, decimal_places=4, max_digits=10,
        help_text=u"utilisez . comme séparateur décimal")
    engagement_oc_enjeu_eau = models.DecimalField(
        u"Surface engagée sur commune à enjeu eau (ha)",
        null=True, blank=True, decimal_places=4, max_digits=10,
        help_text=u"utilisez . comme séparateur décimal")
    installation_bio = models.NullBooleanField(u"installation bio")
    date_engagement_oc = models.DateField(
        u"date d'engagement de l'OC", null=True, blank=True, db_index=True)
    commentaire = models.TextField(null=True, blank=True)

    def __unicode__(self):
        """
        Retourne une représentation unicode de l'objet.
        Attention, problème vicieux: l'admin django appelle cette méthode juste
        après la suppression de l'objet dans la base: il ne faut donc pas
        utiliser %d pour self.id car dans ce cas, self.id vaut None et provoque
        une erreur.
        """
        return u"projet bio de %s (id #%s)" % (self.agriculteur, self.id)

    def clean(self):
#        for attr in (
#            'savoirfaire', 'commercialisation', 'entourage', 'financement',
#            'foncier', 'maindoeuvre', 'materiel', 'motivation', 'reglementation',):
#            if getattr(self, 'pond_' + attr) == -1:
#                break
        if (self.pond_savoirfaire != -1 and self.pond_commercialisation != -1
            and self.pond_entourage != -1 and self.pond_financement != -1
            and self.pond_foncier != -1 and self.pond_maindoeuvre != -1
            and self.pond_materiel != -1 and self.pond_motivation != -1
            and self.pond_reglementation != -1):
            raise ValidationError(msg.ERROR_MESSAGE["projet_sans_contrainte"])
        if self.installation_bio == None:
            raise ValidationError(msg.ERROR_MESSAGE["projet_installation_non_choisi"])

    def agri_id(self):
        if self.agriculteur:
            return self.agriculteur.id
        return u""
    agri_id.short_description = u"ID agriculteur"
    agri_id.verbose_name = u"ID agriculteur"

    def get_absolute_url(self):
        """
        Retourne l'URL associée à l'objet.
        """
        return reverse(
            'admin:{0}_{1}_change'.format(
                self._meta.app_label, self.__class__.__name__.lower()),
            args=[self.pk])


class ProjetBioRo(ProjetBio):
    "La version read-only de ProjetBio"
    class Meta:
        proxy = True
        verbose_name = "Projet de passage en bio (lect. seule)"
        verbose_name_plural = "     Projets de passage en bio (lecture seule)"



class DonneesBio(models.Model):
    "Données de l'observatoire du Gabnor pour une année spécifique."
    class Meta:
        verbose_name = "Informations sur l'exploitation"
        verbose_name_plural = "    Informations sur l'exploitation"
        ordering = ['-annee', 'agriculteur_bio']

    # Champs à inclure dans un export CSV complet
    attrs_export = (
        'agriculteur_bio',
        'agri_id',
        'id',
        'annee',
        'agriculteur_bio__commune__code_commune',
        'agriculteur_bio__commune__arrondissement',
        'agriculteur_bio__commune__enjeu_eau',
        'agriculteur_bio__commune__zone',
        'organisme_certification',
        'aide_conversion_ab',
        'date_aide_conversion_ab',
        'aide_maintien_ab',
        'aide_credit_impot',
        'aide_certification_bio',
        'aide_investissement',
        'aide_installation',
        'aide_mae_t',
        'aide_autres',
        'aide_accompagnement',
        'aide_accompagnement_precisions',
        'biodynamie',
        'etp_associes',
        'etp_familiaux',
        'etp_salaries_permanents',
        'etp_saisonniers',
        'etp_encadrants',
        'personnes_insertion',
        'reseaux_engagements',
        'remarques',
        'tr_date_fin_prevue',
        'tr_repreneur_id',
        'tr_precisions',
        'statut_commercialisation',
        'statut_choisi',
        'lait_vaches',
        'lait_nb_vaches',
        'lait_quota_laiterie',
        'lait_quota_vente_directe',
        'lait_prod_laiterie',
        'lait_prod_vente_directe',
        'lait_laiterie',
        'lait_reformes_cl',
        'lait_taurillons_cl',
        'lait_adultes_cc',
        'lait_veaux_cc',
        'lait_chevres',
        'lait_nb_chevres',
        'lait_chevres_conversion',
        'lait_chevres_prod',
        'bovine',
        'bovine_nb_va',
        'bovine_reformes_cl',
        'bovine_taurillons_cl',
        'bovine_adultes_cc',
        'bovine_veaux_cc',
        'porcs',
        'porcs_nb_truies',
        'porcs_nb_com',
        'moutons',
        'moutons_nb_brebis',
        'moutons_nb_agn_com',
        'animaux_lapins',
        'animaux_volailles',
        'animaux_volailles_chair',
        'animaux_nb_poulets',
        'animaux_nb_lapins',
        'animaux_nb_autres_volailles',
        'animaux_autres_prod',
        'animaux_details_autres_prod',
        'oeufs_nb_pondeuses',
        'oeufs_nb_oeufs',
        'oeufs_acheteur_cl',
        'miel_nb_ruches',
        'formation_annee',
        'formation_annee_laquelle',
        'formation_annee_commentaire',
        'formation_avenir',
        'formation_avenir_type',
        'formation_avenir_details',
    )

    date_collecte = models.DateField(null=True, blank=True)
    date_modification = models.DateField(null=True, blank=True, editable=False)
    date_validation = models.DateField(null=True, blank=True, editable=False)
    technicien = models.ForeignKey(
        auth_models.User, null=True, blank=True, on_delete=models.SET_NULL,
        help_text=u"Personne ayant collecté les données")
    agriculteur_bio = models.ForeignKey(
        'Structure', verbose_name=u"Agriculteur",
        limit_choices_to={'type_pm__gte': 1}
        )
    annee = models.IntegerField(
        u"année", db_index=True, default=datetime.date.today().year)
    organisme_certification = models.ForeignKey(
        ref.OrganismeDeCertification, null=True, blank=True,
        on_delete=models.PROTECT, verbose_name=u"organisme de certification")
    biodynamie = models.NullBooleanField()
    remarques = models.TextField(null=True, blank=True,
            verbose_name=u"remarques de l'équipe")
    # Aides
    aide_conversion_ab = models.NullBooleanField(
            verbose_name=u"aide à la conversion à l'agriculture biologique")
    date_aide_conversion_ab = models.DateField(
        u'date de cette aide', null=True, blank=True,
        help_text=u"En cas d'aide à la conversion, indiquez la date de "
        u"signature du contrat (ou de dépôt pour les MAE)")
    aide_maintien_ab = models.NullBooleanField(
        verbose_name=u"aide au maintien à l'agriculture biologique")
    aide_credit_impot = models.NullBooleanField(
        verbose_name=u"crédit d'impôt bio")
    aide_certification_bio = models.NullBooleanField(
        verbose_name=u"aide à la certification bio")
    aide_investissement = models.NullBooleanField(
        verbose_name=u"aide à l'investissement")
    aide_installation = models.NullBooleanField(
        verbose_name=u"aide à l'installation")
    aide_mae_t = models.NullBooleanField(u"MAE T")
    aide_autres = models.CharField(
        u"autres mesures, précisez", max_length=255, null=True, blank=True)
    aide_accompagnement = models.NullBooleanField(
        u"avez-vous besoin d'un accompagnement pour solliciter une aide ?")
    aide_accompagnement_precisions = models.TextField(
        u"précisions", null=True, blank=True)
    # Main d'oeuvre
    etp_associes = models.DecimalField(
        u"ETP d'associés", null=True, blank=True, max_digits=6,
        decimal_places=2)
    etp_familiaux = models.DecimalField(
        u"ETP familiaux", null=True, blank=True, max_digits=6,
        decimal_places=2)
    etp_salaries_permanents = models.DecimalField(
        u"ETP de salariés permanents", null=True, blank=True, max_digits=6,
        decimal_places=2)
    etp_saisonniers = models.DecimalField(
        u"ETP saisonniers", null=True, blank=True, max_digits=6,
        decimal_places=2)
    etp_encadrants = models.DecimalField(
        u"ETP encadrants", null=True, blank=True, max_digits=6,
        decimal_places=2)
    personnes_insertion = models.IntegerField(
        u"Personnes en contrat d'insertion", null=True, blank=True)
    reseaux_engagements = models.TextField(
        u"Appartenance à des réseaux et engagements", null=True, blank=True)
    # Transmission
    tr_date_fin_prevue = models.DateField(
        u"Date de fin d'activité prévue", null=True, blank=True,
        help_text=u"donnez une date approximative si possible")
    tr_repreneur_id = models.BooleanField(
        u"Repreneur identifié", default=False, blank=True)
    tr_precisions = models.TextField(
        u"Précisions", blank=True, default=u"",
        help_text=u"Identité du repreneur et autre précisions")
    # Commercialisation
    statut_commercialisation = models.NullBooleanField(
        u"utilisation d'un statut spécifique pour la commercialisation")
    statut_choisi = models.ForeignKey(
        ref.StatutCommercialisation, null=True, blank=True,
        verbose_name=u"statut choisi", help_text=u"en cas d'utilisation d'un"
        u" statut spécifique pour la commercialisation, indiquez-le ici.",
        on_delete=models.PROTECT)
    com_ca_total = models.DecimalField(
        u"Chiffre d'affaire total HT", max_digits=10, decimal_places=0,
        null=True, editable=False)
    com_pct_ca_bio = models.DecimalField(
        u"Part du CA bio (%)", max_digits=4, decimal_places=0,
        null=True, editable=False)
    # Accueil
    acc_act_tour = models.NullBooleanField(u"Activité touristique")
    acc_act_peda = models.NullBooleanField(u"Activité pédagogique")
    acc_rest_bio = models.NullBooleanField(u"Restauration bio")
    acc_act_autre = models.NullBooleanField(u"Autres activité")
    detail_activites = models.TextField(
        null=True, blank=True, verbose_name=u"détail des activités")
    # Lait
    lait_vaches = models.NullBooleanField(u"lait de vaches bio")
    lait_nb_vaches = models.PositiveIntegerField(
        u"nombre de vaches laitières", null=True, blank=True)
    lait_quota_laiterie = models.PositiveIntegerField(
        u"quota laiterie (en L)", null=True, blank=True)
    lait_quota_vente_directe = models.PositiveIntegerField(
        "quota vente directe (en L)", null=True, blank=True)
    lait_prod_laiterie = models.PositiveIntegerField(
        u"Lait vendu en laiterie (en L)", null=True, blank=True)
    lait_prod_vente_directe = models.PositiveIntegerField(
        u"Lait vendu en direct (en L)", null=True, blank=True)
    lait_laiterie = models.ForeignKey(
        ref.Laiterie, null=True, blank=True, verbose_name=u"laiterie",
        on_delete=models.PROTECT)
    # viande issue du troupeau laitier
    lait_reformes_cl = models.PositiveIntegerField(
        u"Nb d'animaux adultes type lait valorisés en circuit long (Réformes,"
        u"Génisses, Boeufs)", null=True, blank=True)
    lait_taurillons_cl = models.PositiveIntegerField(
        u"Nb de veaux type lait valorisés en circuit long", null=True,
        blank=True)
    lait_adultes_cc = models.PositiveIntegerField(
        u"nb d'adultes laitiers vendus en circuit court / an", null=True,
        blank=True)
    lait_veaux_cc = models.PositiveIntegerField(
        u"nb de veaux laitiers vendus en circuit court / an", null=True,
        blank=True)
    # Lait de chevre
    lait_chevres = models.NullBooleanField(u"chèvres bio")
    lait_nb_chevres = models.PositiveIntegerField(
        u"nombre de chèvres", null=True, blank=True)
    lait_chevres_conversion = models.PositiveIntegerField(
        u"nombre de chèvres en conversion", null=True, blank=True)
    lait_chevres_prod = models.PositiveIntegerField(
        u"litres de lait de chèvre produits par an", null=True, blank=True)
    # Viande bovine
    bovine = models.NullBooleanField(u"viande bovine bio")
    bovine_nb_va = models.PositiveIntegerField(
        u"nombre VA", null=True, blank=True)
    bovine_reformes_cl = models.PositiveIntegerField(
            null=True, blank=True, verbose_name=
            u"Nb d'animaux adultes type viande valorisés en circuit long",
            help_text=u"(Réformes,Génisses, Boeufs)")
    bovine_taurillons_cl = models.PositiveIntegerField(
            null=True, blank=True,
            verbose_name=u"Nb de veaux type viande valorisés en circuit long")
    bovine_adultes_cc = models.PositiveIntegerField(
            null=True, blank=True, verbose_name=
            u"nb d'adultes race à viande vendus en circuit court / an")
    bovine_veaux_cc = models.PositiveIntegerField(
            null=True, blank=True, verbose_name=
            u"nb de veaux race à viande vendus en circuit court / an")
    # Porcs
    porcs = models.NullBooleanField(u"porc bio")
    porcs_nb_truies = models.PositiveIntegerField(
        u"nombre de truies", null=True, blank=True)
    porcs_nb_com = models.PositiveIntegerField(
        u"nombre de porcs commercialisés",null=True, blank=True)
    # Moutons
    moutons = models.NullBooleanField(u"moutons bio")
    moutons_nb_brebis = models.PositiveIntegerField(
        u"nombre de brebis", null=True, blank=True)
    moutons_nb_agn_com = models.PositiveIntegerField(
        u"nombre d'agneaux commercialisés", null=True, blank=True)
    # Petites animaux
    animaux_lapins = models.NullBooleanField(u"production de lapins bio")
    animaux_volailles = models.NullBooleanField(
        u"production de volailles de chair bio")
    animaux_volailles_chair = models.TextField(
        u"Détail type de volailles (poulets, pintades, dindes)", null=True,
        blank=True)
    animaux_nb_poulets = models.PositiveIntegerField(
        u"nb poulets (produits par an)", null=True, blank=True)
    animaux_nb_lapins = models.PositiveIntegerField(
        u"nb lapins", null=True, blank=True)
    animaux_nb_autres_volailles = models.PositiveIntegerField(
        u"nb autres volailles", null=True, blank=True)
    # Autres viandes
    animaux_autres_prod = models.NullBooleanField(
        u"Autres productions animales bio")
    animaux_details_autres_prod = models.TextField(
        u"détail autres productions animale", null=True, blank=True)
    # Œufs
    oeufs_production_bio = models.NullBooleanField(u"production d'œufs bio")
    oeufs_nb_pondeuses = models.PositiveIntegerField(
        u"nb de pondeuses", null=True, blank=True)
    oeufs_nb_oeufs = models.PositiveIntegerField(
        u"nb d'œufs", null=True, blank=True)
    oeufs_acheteur_cl = models.ForeignKey(
        ref.AcheteurOeufs, null=True, blank=True,
        verbose_name=u"acheteur circuit long", on_delete=models.PROTECT)
    # Miel
    miel_production_bio = models.NullBooleanField(u"apiculture bio")
    miel_nb_ruches = models.PositiveIntegerField(
        u"nb de ruches", null=True, blank=True)
    # Formation
    formation_annee = models.NullBooleanField(
        u"avez-vous participé à une formation au cours de l'année écoulée ?")
    formation_annee_laquelle = models.ForeignKey(
        ref.Formation, null=True, blank=True,
        on_delete=models.PROTECT, verbose_name=u"si oui, laquelle ?")
    formation_annee_commentaire = models.TextField(
        u"commentaires", null=True, blank=True)
    formation_avenir = models.NullBooleanField(
        u"avez-vous des besoins en formation à l'avenir ?")
    formation_avenir_type = models.ForeignKey(
        ref.TypeFormation, null=True, blank=True, on_delete=models.PROTECT,
        verbose_name=u"si oui, lesquelles ?")
    formation_avenir_details = models.TextField(
        u"détails", null=True, blank=True)
    # Infos assolement
    ass_total = models.DecimalField(
        u"Assolement total", blank=True, default=0, max_digits=10,
        decimal_places=4, help_text=
        u"Note: ce total n'est mis à jour qu'après enregistrement de la page")
    ass_total_bio = models.DecimalField(
        u"Assolement total bio", blank=True, default=0, max_digits=10,
        decimal_places=4, help_text=
        u"Note: ce total n'est mis à jour qu'après enregistrement de la page")
    # block A garder pour usage futur 20150907
    lait_vaches_conversion = models.PositiveIntegerField(
        u"nombre de vaches en conversion", null=True, blank=True)
    bovine_nb_conversion = models.PositiveIntegerField(
        u"VA conversion", null=True, blank=True)
    # endblock A garder

    def __unicode__(self):
        "Retourne une représentation unicode de la structure."
        return u"Données de %s pour l'année %s" % (
            self.agriculteur_bio, self.annee)

    def agri_id(self):
        return self.agriculteur_bio.id if self.agriculteur_bio else u""
    agri_id.short_description = u"ID agriculteur bio"
    agri_id.verbose_name = agri_id.short_description

    def get_absolute_url(self):
        "Retourne l'URL du formulaire de modification de l'objet."
        return reverse('admin:%s_%s_change' % (
                self._meta.app_label, self.__class__.__name__.lower()),
                       args=[self.pk])

    def get_assolements(self):
        """
        Retourne une représentation de tous les assolements associés à l'objet
        sous un format adapté à l'export CSV (chaque représentation
        d'assolement est séparé par une nouvelle ligne)
        """
        repr_assolements = []
        for a in self.assolement_set.all():
            repr_assolements.append(a.repr_export())
        return u'\r'.join(repr_assolements)
    get_assolements.short_description = u"assolements"

    def update_assolement(self, commit=False):
        """return change status, ass. total and ass. bio total,
        update field if update=True"""
        total = total_bio = 0
        for ass in self.assolement_set.all():
            sub_total, sub_total_bio = ass.get_total_sau()
            total += sub_total
            total_bio += sub_total_bio
        change = (self.ass_total != total
                  or self.ass_total_bio != total_bio)
        if commit:
            self.ass_total = total
            self.ass_total_bio = total_bio
            self.save()
        return (change, total, total_bio)

    def agri_detail(self):
        if self.agriculteur_bio_id is None:
            return u"-"
        return u'<a href="{0}">{1}</a>'.format(
            self.agriculteur_bio.get_absolute_url(),
            unicode(self.agriculteur_bio))
    agri_detail.short_description = u"Détail agriculteur"
    agri_detail.allow_tags = True

    def clean(self):
        if self.lait_nb_vaches:
            if self.lait_vaches == None:
                raise ValidationError(
                    msg.ERROR_MESSAGE["production_vache_bio_non_choisi"])
        if self.lait_nb_chevres:
            if self.lait_chevres == None:
                raise ValidationError(
                    msg.ERROR_MESSAGE["production_chevre_bio_non_choisi"])
        if (
            self.bovine_nb_va or self.bovine_nb_conversion
            or self.bovine_reformes_cl or self.bovine_taurillons_cl
            or self.bovine_adultes_cc or self.bovine_veaux_cc):
            if self.bovine == None:
                raise ValidationError(
                    msg.ERROR_MESSAGE["production_bovin_bio_non_choisi"])
        if self.porcs_nb_truies or self.porcs_nb_com:
            if self.porcs == None:
                raise ValidationError(
                    msg.ERROR_MESSAGE["production_porc_bio_non_choisi"])
        if self.moutons_nb_brebis:
            if self.moutons == None:
                raise ValidationError(
                    msg.ERROR_MESSAGE["production_mouton_bio_non_choisi"])
        if self.animaux_nb_lapins:
            if self.animaux_lapins == None:
                raise ValidationError(
                    msg.ERROR_MESSAGE["production_lapin_bio_non_choisi"])
        if self.animaux_nb_poulets or self.animaux_nb_autres_volailles:
            if self.animaux_volailles == None:
                raise ValidationError(
                    msg.ERROR_MESSAGE["production_volaille_bio_non_choisi"])
        if self.oeufs_nb_pondeuses:
            if self.oeufs_production_bio == None:
                raise ValidationError(
                    msg.ERROR_MESSAGE["production_oeuf_bio_non_choisi"])
        if self.animaux_details_autres_prod:
            if self.animaux_autres_prod == None:
                raise ValidationError(
                    msg.ERROR_MESSAGE["production_autre_bio_non_choisi"])
        if self.miel_nb_ruches:
            if self.miel_production_bio == None:
                raise ValidationError(
                    msg.ERROR_MESSAGE["production_miel_bio_non_choisi"])


class DonneesBioRo(DonneesBio):
    "La version read-only de DonneesBio"
    class Meta:
        proxy = True
        verbose_name = u"Informations sur l'exploitation (lect. seule)"
        verbose_name_plural = (
            u"    Informations sur l'exploitation (lecture seule)")


class Assolement(models.Model):
    "Informations d'assolement liées à DonneesBio (observatoire Gabnor)."
    attrs_export = (
        'donneesbio__agriculteur_bio',
        ('donneesbio__agriculteur_bio__id', 'ID agriculteur bio'),
        ('donneesbio__id', 'ID donnees' ),
        ('id', 'ID assolement'),
        'culture__categorie',
        'culture',
        'detail_culture',
        'surface_conv',
        'surface_c1',
        'surface_c2',
        'surface_c3',
        'surface_bio',
        'tonnage',
        'prix',
        'client',
        'type_circuit',
    )
    donneesbio = models.ForeignKey(
        'DonneesBio', verbose_name=u"objet donneesbio")
    culture = models.ForeignKey(
        ref.Culture, verbose_name=u"culture", on_delete=models.PROTECT)
    detail_culture = models.CharField(
        u"détail des cultures (si nécessaire)", max_length="255", null=True,
        help_text=u"Complétez si nécessaire la description des cultures",
        blank=True)
    surface_conv = models.DecimalField(
        u"conventionnelle", null=True, blank=True, max_digits=10,
        decimal_places=4)
    surface_c1 = models.DecimalField(
        u"surface C1", null=True, blank=True, max_digits=10, decimal_places=4)
    surface_c2 = models.DecimalField(
        u"surface C2", null=True, blank=True, max_digits=10, decimal_places=4)
    surface_c3 = models.DecimalField(
        u"surface C3", null=True, blank=True, max_digits=10, decimal_places=4)
    surface_bio = models.DecimalField(
        u"surface bio", null=True, blank=True, max_digits=10, decimal_places=4)
    tonnage = models.FloatField(u"tonnage", null=True, blank=True)
    client = models.ForeignKey(
        ref.ClientCulture, null=True, blank=True, verbose_name=u"client",
        on_delete=models.PROTECT)
    prix = models.FloatField(u"Prix", null=True, blank=True)
    type_circuit = models.CharField(
        max_length="10", null=True, blank=True, choices=fch.CHOIX_TYPE_CIRCUIT)

    def get_total_sau(self):
        "return sau and sau bio for this record"
        sub_total = sub_total_bio = 0
        for field in ('conv', 'c1', 'c2', 'c3', 'bio'):
            sub_total += getattr(self, 'surface_' + field, 0) or 0
            if field not in ('conv',):
                sub_total_bio += getattr(self, 'surface_' + field, 0) or 0
        return (sub_total, sub_total_bio)

    def save(self, *args, **kwargs):
        super(Assolement, self).save(*args, **kwargs)
        if not kwargs.pop('raw', False):
            self.donneesbio.update_assolement(commit=True)

    def delete(self, *args, **kwargs):
        super(Assolement, self).delete(*args, **kwargs)
        if not kwargs.pop('raw', False):
            self.donneesbio.update_assolement(commit=True)

    def __unicode__(self):
        "representation unicode de l'objet"
        return u"Assolement de %s" % self.culture

    def repr_export(self):
        """
        Retourne une représentation de l'objet adaptée à l'export CSV
        (permet de faire figurer les assolements dans l'export d'un objet
        donneesbio)
        Note: le séparateur du CSV peut être utilisé car la valeur du champ
        sera entourée de guillements « " ».
        """
        return (u"culture: %s; détail: %s; bio: %s; c1: %s; c2: %s; c3: %s;"
                u" conv: %s; tonnage: %s; client: %s; prix: %s;"
                u" type de circuit: %s"
                % (self.culture, self.detail_culture or '',
                   self.surface_bio or '', self.surface_c1 or '',
                   self.surface_c2 or '', self.surface_c3 or '',
                   self.surface_conv or '',
                   self.tonnage or '', self.client or '',
                   self.prix or '', self.get_type_circuit_display() or ''))


class Transformation(models.Model):
    "sous table des donnees bio"
    class Meta:
        verbose_name=u"Transformation"
        verbose_name_plural=u"Transformations"

    attrs_export = (
        'donneesbio__agriculteur_bio',
        ('donneesbio__agriculteur_bio__id', 'ID agriculteur bio'),
        ('donneesbio__id', 'ID donnees'),
        'ferme',
        'type_pduit',
        'agrmt',
        'list_pduit',
        'part_vol_glob',
        'est_bio',
        'lait_abattoirs_vl',
    )
    donneesbio = models.ForeignKey(
        'DonneesBio', verbose_name=u"objet donneesbio")
    ferme = models.NullBooleanField(u"Transformation à la ferme")
    type_pduit = models.CharField(
        u"Type de produit", max_length=20, choices=fch.TYPE_TRANS)
    agrmt = models.CharField(
        u"Agrément", max_length=20, choices=fch.TYPE_AGREMENT)
    list_pduit = models.TextField(
        u"Détail (liste des produits transformés)", blank=True)
    part_vol_glob = models.DecimalField(
        u"Part du volume global", max_digits=10, decimal_places=0, null=True,
        blank=True, help_text=u"en pourcentage")
    est_bio = models.NullBooleanField(u"Est bio ?")
    lait_abattoirs_vl = models.ManyToManyField(
        ref.Abattoir, null=True, blank=True,
        verbose_name=u"Abattoirs circuit court")

    def __unicode__(self):
        return self.get_type_pduit_display() + u" bio" if self.est_bio else u""


class Commercialisation(models.Model):
    "sous table des donnees bio"
    class Meta:
        verbose_name=u"Commercialisation"
        verbose_name_plural=u"Commercialisations"

    attrs_export = (
        'donneesbio__agriculteur_bio',
        ('donneesbio__agriculteur_bio__id', 'ID agriculteur bio'),
        ('donneesbio__id', 'ID donnees'),
        'type_produit',
        'type_circuit',
        'ch_aff',
        'detail',
        'bio',
    )
    donneesbio = models.ForeignKey(
        'DonneesBio', verbose_name=u"objet donneesbio")
    type_produit = models.CharField(
        u"Type de produit", max_length=20, choices=fch.TYPE_PRODUIT,
        default=fch.TYPE_PRODUIT[0][0])
    type_circuit = models.CharField(
        u"Type de circuit", max_length=20, choices=fch.TYPE_CIRCUIT,
        default=fch.TYPE_CIRCUIT[0][0])
    ch_aff = models.DecimalField(
        u"Chiffre d'affaire HT", max_digits=10, decimal_places=0,
        blank=True, null=True)
    detail = models.TextField(
        blank=True, help_text=u"horaire, date, lieux")
    bio = models.NullBooleanField(u"Est bio ?")

    def get_ch_aff(self):
        if self.ch_aff is None:
            return 0
        return self.ch_aff

    def get_total_ca_for_all_comm(self):
        "compute the total CA and the ratio of bio CA"
        ca_total = 0
        ca_bio = 0
        for comm in self.donneesbio.commercialisation_set.all():
            ca_total += comm.get_ch_aff()
            if comm.bio:
                ca_bio += comm.get_ch_aff()
        return (ca_total, (ca_bio * 100) / ca_total) if ca_total else (0, 0)

    def clean(self, *args, **kwargs):
        if self.bio is None:
            raise ValidationError(msg.ERROR_MESSAGE["commerce_bio_non_choisi"])
        return super(Commercialisation, self).clean(*args, **kwargs)

    def save(self, *args, **kwargs):
        super(Commercialisation, self).save(*args, **kwargs)
        if not kwargs.pop('raw', False):
            self.donneesbio.com_ca_total, self.donneesbio.com_pct_ca_bio = (
                self.get_total_ca_for_all_comm())
            self.donneesbio.save()

    def delete(self, *args, **kwargs):
        super(Commercialisation, self).delete(*args, **kwargs)
        if not kwargs.pop('raw', False):
            self.donneesbio.com_ca_total, self.donneesbio.com_pct_ca_bio = (
                self.get_total_ca_for_all_comm())
            self.donneesbio.save()


class Personne(Contact, models.Model):
    """
    Personne physique dépendant d'une structure.
    .
    Attributs et méthodes hérités de la classe abstraite Contact (hérités aussi
    par Structure) pour permettre le partage du même panier avec les modèles
    basés sur Structures.
     - attrs_publipostage
     - adresse_electronique()
     - get_courriels()
     - liste_contacts_associes()
    """
    class Meta:
        ordering = ['nom', 'prenom', 'structure__designation']
        verbose_name_plural = u"          Personne"

    # Champs à inclure dans un export CSV complet
    attrs_export = [
        'civilite',
        'nom',
        'prenom',
        'date_naissance',
        'date_installation',
        'structure',
        'fonction',
        'adresse',
        'commune',
        'code_postal',
        'courrier_structure',
        'tel_fixe',
        'tel_mobile',
        'fax',
        'courriel',
        'pas_de_courriel',
        'date_creation',
        'date_derniere_modification',
        'reseaux_engagements',
        'utilisateur_derniere_modification',
        ]

    civilite = models.CharField(
        u"civilité", max_length="10", null=True, blank=True,
        choices=fch.CHOIX_CIVILITE)
    nom = models.CharField(max_length=255)
    prenom = models.CharField(max_length=255, verbose_name=u"prénom")
    date_naissance = models.DateField(
        u"date de naissance", null=True, blank=True)
    structure = models.ForeignKey("Structure", verbose_name=u"affilié à")
    date_installation = models.DateField(
        u"date d'installation", null=True, blank=True)
    fonction = models.CharField(
        u"fonction", max_length="255", null=True, blank=True)
    adresse = models.TextField(
        u"adresse personnelle", blank=True,
        help_text="Saisissez l'adresse de la personne, sur 4 lignes maximum.")
    commune = models.ForeignKey(
        ref.Commune, null=True, blank=True, on_delete=models.PROTECT)
    # Certaines communes peuvent avoir plusieurs codes postaux
    code_postal = models.CharField(
        max_length=10, null=True, blank=True,
        help_text=u"Laisser vide pour utiliser celui associé à la commune")
    # la ville contient également le code cedex (à utiliser pour l'adresse)
    ville = models.CharField(
        max_length="255", null=True, blank=True, help_text=u"Nom de ville"
        u" avec code cédex si nécessaire, si différent du nom de la commune.")
    courrier_structure = models.BooleanField(
        u"Recevoir les courriers addressés à la structure", default=False,
        help_text=u"Cochez cette case si la personne doit recevoir les "
        u"courriers papier envoyés à la structure.")
    tel_fixe = models.CharField(u"téléphone fixe", max_length=100, blank=True)
    tel_mobile = models.CharField(
        u"téléphone mobile", max_length=100, blank=True)
    fax = models.CharField(max_length=100, null=True, blank=True)
    courriel = models.CharField(max_length=255, blank=True)
    pas_de_courriel = models.BooleanField(
        u"pas d'envoi automatique de courriel", default=False,
        help_text=u"Cochez cette case si la personne ne souhaite"
        u" pas recevoir de courriel envoyé automatiquement.")
    reseaux_engagements = models.TextField(
        u"Appartenance à des réseaux et engagements", null=True, blank=True)
    date_creation = models.DateField(
        u"date de saisie de l'enregistrement", editable=False)
    date_derniere_modification = models.DateField(
        u"date de dernière modification de informations", editable=False)
    utilisateur_derniere_modification = models.ForeignKey(
        auth_models.User, editable=False, on_delete=models.SET(get_rem_user),
        verbose_name=u"dernière modification par")

    def __unicode__(self):
        """
        Retourne une représentation unicode de la structure.
        """
        if not self.nom and not self.prenom:
            return u"Personne non nommée #%s" % self.id
        return u"%s %s" % (self.prenom, self.nom)

    def get_absolute_url(self):
        """
        Retourne l'URl associée à l'objet.
        """
        return reverse('admin:%s_%s_change' \
                % (self._meta.app_label, self.__class__.__name__.lower()),
                args =[self.pk])

    def nom_prenom(self):
        """
        Retourne les nom et prénom de la personne sous forme de chaîne de
        caractères.
        """
        return u"%s %s" % (self.nom.upper(), self.prenom)

    def contenu_etiquette(self):
        """
        Retourne le contenu à afficher dans l'étiquette sous forme de liste
        dont chaque élément représente une ligne.
        """
        contenu = []
        # On doit remplacer & par &amp; à cause d'un problème dans reportlab
        # qui génère une erreur XML lorsque le nom contient un "&"
        contenu.append(self.nom_prenom().replace("&", "&amp;"))
        contenu.extend(self.adresse.split('\n'))
        contenu.append("%s %s" % (self.code_postal or '', self.ville or ''))
        return contenu

    def adresse_electronique(self):
        """
        Retourne une liste de 2 éléments:
        - le nom de la structure
        - l'adresses email associée
        Retourne une liste vide si aucune adresse courriel n'est renseignée.
        """
        if not self.courriel:
            return ()
        if self.nom:
            nom = u"%s %s" % (self.prenom, self.nom)
        else:
            nom = u""
        return (nom, self.courriel)

    def get_courriels(self):
        """
        Méthode commune aux structures et personnes: retourne la liste des
        adresses emails (dans Personne il n'y en a qu'une)
        """
        adr = self.adresse_electronique()
        if adr and not self.pas_de_courriel:
            return [ adr ]
        return []

    def clean(self):
        if self.structure.type_pm >= 1 and not self.date_naissance:
            raise ValidationError(msg.ERROR_MESSAGE["pers_af_agri_sans_date_naiss"])

    def save(self, *args, **kwargs):
        """
        Remplit les champs date_creation et date_derniere_modification
        manuellement (la fonctionnalité auto_now de django est douteuse,
        mieux vaut le gérer manuellement).
        """
        if self.id is None:
            self.date_creation = datetime.date.today()
        self.date_derniere_modification = datetime.date.today()
        super(Personne, self).save(*args, **kwargs)


class PersonneRo(Personne):
    "La version read-only de personne"
    class Meta:
        proxy = True
        verbose_name = "Personne (lecture seule)"
        verbose_name_plural = "         Personnes (lecture seule)"


### Classes secondaires
class RdvPersonne(models.Model):
    "modele intermediaire: participation d'une personne a un evt"
    class Meta:
        db_table = "contacts_rdv_personnes"
        unique_together = ('rdv', 'personne')

    rdv = models.ForeignKey("Rdv")
    personne = models.ForeignKey(
        "Personne", verbose_name=u"Personne physique", help_text=
        u"pour les rendez vous collectifs")
        
    def is_organisateur(self):
        if (self.personne.structure):
           return  u'<a href="{0}">{1}</a>'.format(
            self.personne.structure.get_absolute_url(),
            unicode(self.personne.structure))
        return ""
    is_organisateur.short_description = "Organisation"
    is_organisateur.allow_tags = True
    
    def __unicode__(self):
        return u"Participation de {0} à {1}".format(self.personne, self.rdv)


class RdvPersonneMorale(models.Model):
    "modele intermediaire: participation d'une structure a un rdv"
    class Meta:
        db_table = "contacts_rdv_personnes_morales"
        unique_together = ('rdv', 'structure')

    rdv = models.ForeignKey("Rdv")
    structure = models.ForeignKey(
        "Structure", verbose_name=u"Personne morale", help_text=
        u"pour les rendez vous collectifs")        
    def __unicode__(self):
        return u"Participation de {0} à {1}".format(self.structure, self.rdv)


class Rdv(models.Model):
    "Rendez vous generiques"
    class Meta:
        verbose_name = u"Rendez vous ou Évenement"
        verbose_name_plural = u"Rendez vous et Évenements"
        ordering = ['-date']
        get_latest_by = 'date'

    attrs_export = (
        'proj_id',
        'personne_morale__id',
        'personne_morale__designation',
        'personne_morale__get_type_pm_display',
        'projet_bio__technicien',
        'personne_morale__adresse',
        'personne_morale__commune__code_postal',
        'personne_morale__commune__nom',
        'personne_morale__commune__code_commune',
        'personne_morale__commune__arrondissement',
        'personne_morale__commune__enjeu_eau',
        'personne_morale__etiquettes',
        'personne_morale__type_activite',
        'personne_morale__qual_compl',
        'personne_morale__type_activite_pri',
        'personne_morale__qual_compl_conv',
        'personne_morale__last_donneesbio__ass_total',
        'personne_morale__last_donneesbio__ass_total_bio',
        'personne_morale__sau_eau_bio',
        'personne_morale__sau_eau',
        'date',
        'technicien',
        'get_type_rdv_display',
        'commentaire',
        'acc_duree',
        'acc_distance',
        'projet_bio__description',
        'projet_bio__freins_identifies',
        'projet_bio__pond_savoirfaire',
        'projet_bio__pond_motivation',
        'projet_bio__pond_entourage',
        'projet_bio__pond_foncier',
        'projet_bio__pond_materiel',
        'projet_bio__pond_reglementation',
        'projet_bio__pond_commercialisation',
        'projet_bio__pond_financement',
        'projet_bio__pond_maindoeuvre',
        'projet_bio__suites_a_donner',
        'projet_bio__remarques',
        'projet_bio__date_probable_certification',
        'projet_bio__date_certification',
        'projet_bio__engagement_surf',
        'projet_bio__engagement_oc_enjeu_eau',
        'projet_bio__installation_bio',
        'projet_bio__date_engagement_oc',
        'projet_bio__commentaire',
        #'personnes_morales__count',
        'personnes__count',
        'titre_evt',
        'lieu',
        'finance_prog',
        'est_confirme',
        'zone_cible',
        'public_cible',
        'organisateur',
        'intervenant',
        'objectif',
        'res_attendu',
        )

    date = models.DateField(db_index=True)
    technicien = models.ForeignKey(
        auth_models.User, on_delete=models.SET(get_rem_user))
    type_rdv = models.CharField(
        max_length=25, choices=fch.CHOIX_TYPE_RDV,
        verbose_name=u"type")
    projet_bio = models.ForeignKey(
        'ProjetBio', blank=True, null=True, verbose_name=u"Projet bio",
        on_delete=models.SET_NULL)
    titre_evt = models.ForeignKey(
        'references.Formation', blank=True, null=True,
        verbose_name=u"Désignation évenement", on_delete=models.SET_NULL)
    personne_morale = models.ForeignKey(
        'Structure', related_name="rdv_ind_set", blank=True, null=True,
        verbose_name="Personne morale", help_text=
        u"pour les rendez vous individuels")
    personnes_morales = models.ManyToManyField(
            'Structure', related_name="evt_set", through="RdvPersonneMorale", verbose_name=
            u"Participants", help_text=u"pour les rendez vous collectifs")
    personnes = models.ManyToManyField(
        'Personne', through="RdvPersonne", verbose_name=u"Participant(s) ind.")
    zone_cible = models.ManyToManyField(
        'references.Zone', null=True, blank=True)
    public_cible = models.ManyToManyField(
        'references.PublicCible', null=True, blank=True)
    organisateur = models.ForeignKey(
        Personne, null=True, blank=True, on_delete=models.PROTECT,
        related_name="organisation_pers")
    intervenant = models.ForeignKey(
        Personne, null=True, blank=True, on_delete=models.PROTECT,
        related_name="intervenant_pers")
    objectif = models.TextField(blank=True, default="")
    res_attendu = models.TextField(blank=True, default="")
    lieu = models.ForeignKey(
        ref.Commune, null=True, blank=True, on_delete=models.SET_NULL,
        help_text=u"si vide, rempli automatiquement", )
    # accounting: duree et distance associees au rendez vous
    acc_duree = models.DecimalField(
        u"Temps passé", max_digits=3, decimal_places=1, null=True, blank=True,
        help_text=u"en heure")
    acc_distance = models.DecimalField(
        u"Distance associée", max_digits=4, decimal_places=1, null=True,
        blank=True, help_text=u"en km")
    finance_prog = models.ForeignKey(
        ref.FinanceProg, null=True, blank=True, on_delete=models.SET_NULL,
        verbose_name=u"Financement / Programme concerné")
    est_confirme = models.BooleanField(u"est confirmé", blank=True)
    commentaire = models.TextField(blank=True)
    pk_rdv_suivi = models.IntegerField(default=0, editable=False)
    pk_rdv_bio = models.IntegerField(default=0, editable=False)

    def struct_af_item(self, item):
        if (getattr(self, item) is None
            or getattr(self, item).structure is None):
            return 'N/A'
        return '<a href="{0}" target="_BLANK">{1}</a>'.format(
            getattr(self, item).structure.get_absolute_url(),
            getattr(self, item).structure)

    def struct_af_org(self):
        return self.struct_af_item('organisateur')
    struct_af_org.short_description = "P. morale affiliée à l'organisation"
    struct_af_org.allow_tags = True

    def struct_af_int(self):
        return self.struct_af_item('intervenant')
    struct_af_int.short_description = "P. morale affiliée à l'intervenant"
    struct_af_int.allow_tags = True

    def proj_id(self):
        "retoune l'identifiant du projet"
        return self.projet_bio.id if self.projet_bio else u""
    proj_id.short_description = u"ID Projet"

    def get_cat(self):
        "retourne la categorie de rendez vous (indiv, collect)"
        resp = set(
            key for key, value in fch.MTX_TYPE_RDV.iteritems()
            if self.type_rdv in value)
        return resp

    def get_absolute_url(self):
        return reverse(
            'admin:{0}_{1}_change'.format(
                self._meta.app_label, self.__class__.__name__.lower()),
            args=[self.pk])

    def save(self, *args, **kwargs):
        if self.lieu is None:
            pers_morale = self.rdvpersonnemorale_set.all()[:1]
            if len(pers_morale) == 1:
                pers_morale = pers_morale[0].structure
                self.lieu = pers_morale.commune
        super(Rdv, self).save(*args, **kwargs)

    def clean(self):
        cat = self.get_cat()
        if self.est_confirme:
            if 'pbio' in cat and not self.projet_bio:
                raise ValidationError(msg.ERROR_MESSAGE["rdv_pbio_sans_pbio"])
            if 'indiv' in cat and not self.personne_morale:
                raise ValidationError(msg.ERROR_MESSAGE["rdv_indiv_sans_part"])
            if 'collect' in cat:
                if not self.titre_evt:
                    raise ValidationError(
                        msg.ERROR_MESSAGE["rdv_collec_sans_titre"])

    def __unicode__(self):
        return u"{0}-{1}-{2}".format(
            self.date.strftime(LOCAL_DATE_FORMAT).decode('utf-8'),
            self.get_type_rdv_display(), self.technicien)


class RdvRo(Rdv):
    "La version read-only de rdv"
    class Meta:
        proxy = True
        verbose_name = "Rendez vous ou Évenement (lect. seule)"
        verbose_name_plural = "Rendez vous et Évenements (lecture seule)"



class AdhesionGabnor(models.Model):
    "Adhésion Gabnor."
    class Meta:
        ordering = ['-date']
        get_latest_by = 'date'
        verbose_name = "adhésion GABNOR"
        verbose_name_plural = "      Adhésions GABNOR"

    # Champs à inclure dans un export CSV complet
    attrs_export = [ 'agriculteur_bio', 'date', 'montant', ]

    # todo: renommer agriculteur_bio en agriculteur
    agriculteur_bio = models.ForeignKey(
        'Structure', limit_choices_to={'type_pm__gte': 1})
    date = models.DateField(db_index=True)
    montant = models.DecimalField(
        u"montant de l'adhésion (en euros)", max_digits=8, decimal_places=2,
        null=True, blank=True)
    annee = models.IntegerField(
        u"Adhésion de l'année", null=True, blank=True,
        help_text=u"Laisser vide pour utiliser l'année de la date d'adhésion")

    def __unicode__(self):
        """
        Représentation unicode de l'objet (utilisé notamment pour affichage
        dans l'objet AgriculteurBio).
        """
        return (u"Adhésion du %s de %.2f EUR de %s"
                % (self.date.strftime(LOCAL_DATE_FORMAT).decode('utf8'),
                   self.montant, self.agriculteur_bio))

    def save(self, *args, **kwargs):
        """
        Surchargé pour mettre à jour les dates d'adhésions de l'agriculteur.
        """
        if self.annee is None:
            self.annee = self.date.year
        super(AdhesionGabnor, self).save(*args, **kwargs)
        self._update_dates_agriculteurbio()

    def delete(self, *args, **kwargs):
        """
        Surchargé pour mettre à jour les dates d'adhésions de l'agriculteur.
        (cette méthode n'est pas appelée pour les suppressions en masse,
        mais ça n'arrivera pas dans un usage normal de l'application).
        """
        super(AdhesionGabnor, self).delete(*args, **kwargs)
        self._update_dates_agriculteurbio()

    def _update_dates_agriculteurbio(self):
        """
        Calcule et sauvegarde les dates de première et dernière adhésion,
        repectivement date_adhesion et date_renouvellement dans l'objet
        AgriculteurBio.
        """
        adhesions_qs = (AdhesionGabnor.objects
                        .filter(agriculteur_bio=self.agriculteur_bio)
                        .order_by('date'))
        if len(adhesions_qs) != 0:
            premiere_adhesion = adhesions_qs[0]
        else:
            premiere_adhesion = None
        if len(adhesions_qs) > 1:
            # le renouvellement est la dernière adhésion en date
            renouvellemenent = adhesions_qs[len(adhesions_qs)-1]
        else:
            renouvellemenent = None
        self.agriculteur_bio.premiere_adhesion = premiere_adhesion
        self.agriculteur_bio.renouvellement_adhesion = renouvellemenent
        self.agriculteur_bio.save()


class AdhesionGabnorRo(AdhesionGabnor):
    "La version read-only de AdhesionGabnor"
    class Meta:
        proxy = True
        verbose_name = "adhésion GABNOR (lecture seule)"
        verbose_name_plural = "      Adhésions GABNOR (lecture seule)"


class DocumentJoint(models.Model):
    "Documents joints aux rendez-vous."
    class Meta:
        verbose_name = "document joint"
        verbose_name_plural = "Documents Joints"

    document =  models.FileField(upload_to="pj/%Y/%m/%d/")
    date_depot = models.DateTimeField(auto_now=True)
    rdv = models.ForeignKey(Rdv, null=True, blank=True)

    def __unicode__(self):
        "Retourne une représentation unicode de l'objet."
        return u"%s : %s" % (self.document, self.date_depot)
# EOF
